@@ -21,96 +21,91 @@
 from sys import platform
 import os
 import logging
-from os.path import split
 import re
 import json
 from copy import deepcopy
-import scipy as s
-from scipy.stats import norm as normal
-from scipy.interpolate import interp1d
+import numpy as np
+import scipy.stats
+import scipy.interpolate
 
 from ..radiative_transfer.look_up_tables import TabularRT, FileExistsError
 from ..core.common import json_load_ascii, recursive_replace
 
-from ..core.common import combos, VectorInterpolator, eps, load_wavelen
+from ..core.common import VectorInterpolator
+from isofit.configs.sections.radiative_transfer_config import RadiativeTransferEngineConfig
+from isofit.configs import Config
+from isofit.core.geometry import Geometry
+import subprocess
 
 
 ### Variables ###
 
 eps = 1e-5  # used for finite difference derivative calculations
 
-modtran_bands_available = ['modtran_vswir', 'modtran_tir']
-
 ### Classes ###
 
 
 class ModtranRT(TabularRT):
     """A model of photon transport including the atmosphere."""
 
-    def __init__(self, band_mode_string, config, statevector_names):
+    def __init__(self, engine_config: RadiativeTransferEngineConfig, full_config: Config):
         """."""
 
-        TabularRT.__init__(self, config)
+        # Specify which of the potential MODTRAN LUT parameters are angular, which will be handled differently
+        self.angular_lut_keys_degrees = ['OBSZEN', 'TRUEAZ', 'viewzen', 'viewaz', 'solzen', 'solaz']
+        self.angular_lut_keys_radians = []
 
-        self.band_mode_string = band_mode_string
-        if self.band_mode_string not in modtran_bands_available:
-            raise NotImplementedError
+        super().__init__(engine_config, full_config)
 
-        self.modtran_dir = self.find_basedir(config)
-        flt_name = 'wavelengths_' + self.band_mode_string + '.flt'
+        # Flag to determine if MODTRAN should operate with reflectivity = 1
+        # (enabling thermal_upwelling and thermal_downwelling to be determined - see comments below)
+        self.treat_as_emissive = False
+        if self.wl[0] > 2500:
+            self.treat_as_emissive = True
+
+        self.modtran_dir = self.find_basedir(engine_config)
+        flt_name = 'wavelengths_{}_{}_{}.flt'.format(
+            engine_config.engine_name, self.wl[0], self.wl[-1])
         self.filtpath = os.path.join(self.lut_dir, flt_name)
-        self.template = deepcopy(json_load_ascii(
-            config['modtran_template_file'])['MODTRAN'])
+        self.template = deepcopy(json_load_ascii(engine_config.template_file)['MODTRAN'])
 
         # Insert aerosol templates, if specified
-        if 'aerosol_template_file' in config:
+        if engine_config.aerosol_model_file is not None:
             self.template[0]['MODTRANINPUT']['AEROSOLS'] = \
-                deepcopy(json_load_ascii(config['aerosol_template_file']))
+                deepcopy(json_load_ascii(engine_config.aerosol_template_file))
 
         # Insert aerosol data, if specified
-        if 'aerosol_model_file' in config:
-            aer_data = s.loadtxt(config['aerosol_model_file'])
+        if engine_config.aerosol_model_file is not None:
+            aer_data = np.loadtxt(engine_config.aerosol_model_file)
             self.aer_wl = aer_data[:, 0]
-            aer_data = aer_data[:, 1:].T
+            aer_data = np.transpose(aer_data[:, 1:])
             self.naer = int(len(aer_data)/3)
             aer_absc, aer_extc, aer_asym = [], [], []
             for i in range(self.naer):
                 aer_extc.append(aer_data[i*3])
                 aer_absc.append(aer_data[i*3+1])
                 aer_asym.append(aer_data[i*3+2])
-            self.aer_absc = s.array(aer_absc)
-            self.aer_extc = s.array(aer_extc)
-            self.aer_asym = s.array(aer_asym)
+            self.aer_absc = np.array(aer_absc)
+            self.aer_extc = np.array(aer_extc)
+            self.aer_asym = np.array(aer_asym)
 
-        if self.band_mode_string.lower() == 'modtran_vswir':
-            self.modtran_lut_names = ['rhoatm', 'transm', 'sphalb', 'transup']
-        elif self.band_mode_string.lower() == 'modtran_tir':
-            self.modtran_lut_names = ['thermal_upwelling', 
-                'thermal_downwelling', 'rhoatm', 'transm', 'sphalb', 'transup']
+        self.modtran_lut_names = ['rhoatm', 'transm', 'sphalb', 'transup']
+        if self.treat_as_emissive:
+            self.modtran_lut_names = ['thermal_upwelling',
+                                      'thermal_downwelling'] + self.modtran_lut_names
 
-        self.angular_lut_keys_degrees = ['OBSZEN','TRUEAZ','viewzen','viewaz',
-            'solzen','solaz']
+        self.last_point_looked_up = np.zeros(self.n_point)
+        self.last_point_lookup_values = np.zeros(self.n_point)
 
         # Build the lookup table
         self.build_lut()
 
-        # This array is used to handle the potential indexing mismatch between 
-        # the 'global statevector' (which may couple multiple radiative transform 
-        # models) and this statevector. It should never be modified
-        full_to_local_statevector_position_mapping = []
-        for sn in self.statevec:
-            ix = statevector_names.index(sn)
-            full_to_local_statevector_position_mapping.append(ix)
-        self._full_to_local_statevector_position_mapping = \
-            s.array(full_to_local_statevector_position_mapping)
-
     def find_basedir(self, config):
         """Seek out a modtran base directory."""
 
-        try:
-            return config['modtran_directory']
-        except KeyError:
-            pass  # fall back to environment variable
+        if config.engine_base_dir is not None:
+            return config.engine_base_dir
+
         try:
             return os.getenv('MODTRAN_DIR')
         except KeyError:
@@ -129,7 +124,7 @@ class ModtranRT(TabularRT):
                     lines.append(f.readline())
                 except UnicodeDecodeError:
                     pass
-            #lines = f.readlines()
+
             for i, line in enumerate(lines):
                 if "SINGLE SCATTER SOLAR" in line:
                     ts = i+5
@@ -139,30 +134,32 @@ class ModtranRT(TabularRT):
             if ts < 0:
                 logging.error('%s is missing solar geometry' % infile)
                 raise ValueError('%s is missing solar geometry' % infile)
-        szen = s.array([float(lines[i].split()[3])
-                        for i in range(ts, te)]).mean()
+        szen = np.array([float(lines[i].split()[3])
+                         for i in range(ts, te)]).mean()
         return szen
 
     def load_chn(self, infile, coszen):
         """Load a '.chn' output file and parse critical coefficient vectors. 
 
            These are:
-             wl      - wavelength vector
-             sol_irr - solar irradiance
-             sphalb  - spherical sky albedo at surface
-             transm  - diffuse and direct irradiance along the 
+             * wl      - wavelength vector
+             * sol_irr - solar irradiance
+             * sphalb  - spherical sky albedo at surface
+             * transm  - diffuse and direct irradiance along the
                           sun-ground-sensor path
-             transup - transmission along the ground-sensor path only 
+             * transup - transmission along the ground-sensor path only
 
-            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
-             Be careful with these! They are to be used only by the 
-             modtran_tir functions because MODTRAN must be run with a 
-             reflectivity of 1 for them to be used in the RTM defined
-             in radiative_transfer.py.
-             thermal_upwelling - atmospheric path radiance
-             thermal_downwelling - sky-integrated thermal path radiance 
+           !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
+            Be careful with these! They are to be used only by the
+            modtran_tir functions because MODTRAN must be run with a
+            reflectivity of 1 for them to be used in the RTM defined
+            in radiative_transfer.py.
+
+            * thermal_upwelling - atmospheric path radiance
+            * thermal_downwelling - sky-integrated thermal path radiance
                 reflected off the ground and back into the sensor.
-            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
+
+           !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 
            We parse them one wavelength at a time."""
 
@@ -178,9 +175,9 @@ class ModtranRT(TabularRT):
                 toks = re.findall(r"[\S]+", line.strip())
                 wl, wid = float(toks[0]), float(toks[8])  # nm
                 solar_irr = float(toks[18]) * 1e6 * \
-                    s.pi / wid / coszen  # uW/nm/sr/cm2
+                    np.pi / wid / coszen  # uW/nm/sr/cm2
                 rdnatm = float(toks[4]) * 1e6  # uW/nm/sr/cm2
-                rhoatm = rdnatm * s.pi / (solar_irr * coszen)
+                rhoatm = rdnatm * np.pi / (solar_irr * coszen)
                 sphalb = float(toks[23])
                 transm = float(toks[22]) + float(toks[21])
                 transup = float(toks[24])
@@ -189,11 +186,11 @@ class ModtranRT(TabularRT):
                 thermal_emission = float(toks[11])
                 thermal_scatter = float(toks[12])
                 thermal_upwelling = (thermal_emission + thermal_scatter) / \
-                                    wid * 1e6  # uW/nm/sr/cm2
+                    wid * 1e6  # uW/nm/sr/cm2
 
                 # Be careful with these! See note in function comments above
                 # grnd_rflt already includes ground-to-sensor transmission
-                grnd_rflt = float(toks[16]) 
+                grnd_rflt = float(toks[16])
                 thermal_downwelling = grnd_rflt / wid * 1e6  # uW/nm/sr/cm2
 
                 sols.append(solar_irr)
@@ -206,14 +203,14 @@ class ModtranRT(TabularRT):
                 thermal_downwellings.append(thermal_downwelling)
 
                 wls.append(wl)
-        params = [s.array(i) for i in [wls, sols, rhoatms, transms, sphalbs,
-                    transups, thermal_upwellings, thermal_downwellings]]
+        params = [np.array(i) for i in [wls, sols, rhoatms, transms, sphalbs,
+                                        transups, thermal_upwellings, thermal_downwellings]]
         return tuple(params)
 
     def ext550_to_vis(self, ext550):
         """."""
 
-        return s.log(50.0) / (ext550 + 0.01159)
+        return np.log(50.0) / (ext550 + 0.01159)
 
     def modtran_driver(self, overrides):
         """Write a MODTRAN 6.0 input file."""
@@ -221,7 +218,7 @@ class ModtranRT(TabularRT):
         param = deepcopy(self.template)
 
         if hasattr(self, 'aer_absc'):
-            fracs = s.zeros((self.naer))
+            fracs = np.zeros((self.naer))
 
         # Perform overrides
         for key, val in overrides.items():
@@ -253,7 +250,7 @@ class ModtranRT(TabularRT):
             total_asym = self.aer_asym.T.dot(norm_fracs)
 
             # Normalize to 550 nm
-            total_extc550 = interp1d(self.aer_wl, total_extc)(0.55)
+            total_extc550 = scipy.interpolate.interp1d(self.aer_wl, total_extc)(0.55)
             lvl0 = param[0]['MODTRANINPUT']['AEROSOLS']['IREGSPC'][0]
             lvl0['NARSPC'] = len(self.aer_wl)
             lvl0['VARSPC'] = [float(v) for v in self.aer_wl]
@@ -279,28 +276,77 @@ class ModtranRT(TabularRT):
         if not os.path.exists(self.filtpath):
             self.wl2flt(self.wl, self.fwhm, self.filtpath)
 
+        # Check that the H2OSTR value, if present, is not too high.
+        # MODTRAN caps the value at 5x profile specified value or 100% RH, as
+        # defined in PDF-page 52 of the MODTRAN user guide.
+        if 'H2OSTR' in self.lut_names:
+            if 'H2OOPT' in self.template[0]['MODTRANINPUT']['ATMOSPHERE'].keys() and self.template[0]['MODTRANINPUT']['ATMOSPHERE']['H2OOPT'] == '+':
+                logging.info('H2OOPT found in MODTRAN template - ignoring H2O upper bound')
+            else:
+                # Only do this check if we don't have a LUT provided:
+                need_to_rebuild = np.any([not self.required_results_exist(x) for x in self.get_lut_filenames()])
+                if need_to_rebuild:
+
+                    # Define a realistic point, based on lut grid
+                    point = np.array([x[-1] for x in self.lut_grids])
+
+                    # Set the H2OSTR value as arbitrarily high - 50 g/cm2 in this case
+                    point[self.lut_names.index('H2OSTR')] = 50
+
+                    filebase = os.path.join(os.path.dirname(self.files[-1]), 'H2O_bound_test')
+                    cmd = self.rebuild_cmd(point, filebase)
+
+                    # Run MODTRAN for up to 10 seconds - this should be plenty of time
+                    cwd = os.getcwd()
+                    if os.path.isdir(self.lut_dir) is False:
+                        os.mkdir(self.lut_dir)
+                    os.chdir(self.lut_dir)
+                    try:
+                        subprocess.call(cmd, shell=True, timeout=10)
+                    except:
+                        pass
+                    os.chdir(cwd)
+
+
+                    max_water = None
+                    with open(os.path.join(self.lut_dir,filebase + '.tp6'), errors='ignore') as tp6file:
+                        for count, line in enumerate(tp6file):
+                            if 'The water column is being set to the maximum' in line:
+                                max_water = line.split(',')[1].strip()
+                                max_water = float(max_water.split(' ')[0])
+                                break
+
+                    if max_water is None:
+                        logging.error('Could not find MODTRAN H2O upper bound in file {}'.format(filebase + '.tp6'))
+                        raise KeyError('Could not find MODTRAN H2O upper bound')
+
+                    if np.max(self.lut_grids[self.lut_names.index('H2OSTR')]) > max_water:
+                        logging.error('MODTRAN max H2OSTR with current profile is {}, while H2O lut_grid is {}.  Either adjust MODTRAN profile or lut_grid.  To over-ride MODTRANs maximum allowable value, set H2OOPT to "+"'.format(max_water, self.lut_grids[self.lut_names.index('H2OSTR')]))
+                        raise KeyError('MODTRAN H2O lut grid is invalid - see logs for details.')
+
+
         TabularRT.build_lut(self, rebuild)
 
         mod_outputs = []
         for point, fn in zip(self.points, self.files):
-            chnfile = self.lut_dir+'/'+fn+'.chn'
             mod_outputs.append(self.load_rt(fn))
 
         self.wl = mod_outputs[0]['wl']
         self.solar_irr = mod_outputs[0]['sol']
-        self.coszen = s.cos(mod_outputs[0]['solzen'] * s.pi / 180.0)
+        self.coszen = np.cos(mod_outputs[0]['solzen'] * np.pi / 180.0)
 
         dims_aug = self.lut_dims + [self.n_chan]
         for key in self.modtran_lut_names:
-            temp = s.zeros(dims_aug, dtype=float)
+            temp = np.zeros(dims_aug, dtype=float)
             for mod_output, point in zip(mod_outputs, self.points):
-                ind = [s.where(g == p)[0] for g, p in \
-                    zip(self.lut_grids, point)]
+                ind = [np.where(g == p)[0] for g, p in
+                       zip(self.lut_grids, point)]
                 ind = tuple(ind)
                 temp[ind] = mod_output[key]
 
-            self.luts[key] = VectorInterpolator(self.lut_grids, temp, 
-                    self.lut_interp_types)
+            self.luts[key] = VectorInterpolator(self.lut_grids, temp,
+                                                self.lut_interp_types)
+
 
     def rebuild_cmd(self, point, fn):
         """."""
@@ -317,13 +363,9 @@ class ModtranRT(TabularRT):
 
         # Check rebuild conditions: LUT is missing or from a different config
         infilename = 'LUT_'+fn+'.json'
-        infilepath = os.path.join(self.lut_dir, infilename)
-        outchnname = os.path.join(self.lut_dir, fn+'.chn')
-        outtp6name = os.path.join(self.lut_dir, fn+'.tp6')
+        infilepath = os.path.join(self.lut_dir, 'LUT_'+fn+'.json')
 
-        if not os.path.exists(infilepath) or\
-           not os.path.exists(outchnname) or\
-           not os.path.exists(outtp6name):
+        if not self.required_results_exist(fn):
             rebuild = True
         else:
             # We compare the two configuration files, ignoring names and
@@ -335,7 +377,7 @@ class ModtranRT(TabularRT):
                 current_config[0]['MODTRANINPUT']['SPECTRAL']['FILTNM'] = ''
                 modtran_config[0]['MODTRANINPUT']['SPECTRAL']['FILTNM'] = ''
                 current_str = json.dumps(current_config)
-                modtran_str = json.dumps(current_config)
+                modtran_str = json.dumps(modtran_config)
                 rebuild = (modtran_str.strip() != current_str.strip())
 
         if not rebuild:
@@ -359,27 +401,37 @@ class ModtranRT(TabularRT):
             raise SystemExit("MODTRAN directory not defined in config file.")
 
         # Generate the CLI path
-        cmd = self.modtran_dir+'/bin/'+xdir[platform]+'/mod6c_cons '+infilename
+        cmd = os.path.join(self.modtran_dir, 'bin', xdir[platform], 'mod6c_cons ' + infilename)
         return cmd
 
+    def required_results_exist(self, fn):
+        infilename = os.path.join(self.lut_dir, 'LUT_'+fn+'.json')
+        outchnname = os.path.join(self.lut_dir, fn+'.chn')
+        outtp6name = os.path.join(self.lut_dir, fn+'.tp6')
+
+        if os.path.isfile(infilename) and os.path.isfile(outchnname) and os.path.isfile(outtp6name):
+            return True
+        else:
+            return False
+
     def load_rt(self, fn):
         """."""
 
-        tp6file = self.lut_dir+'/'+fn+'.tp6'
+        tp6file = os.path.join(self.lut_dir, fn+'.tp6')
         solzen = self.load_tp6(tp6file)
-        coszen = s.cos(solzen * s.pi / 180.0)
+        coszen = np.cos(solzen * np.pi / 180.0)
 
-        chnfile = self.lut_dir+'/'+fn+'.chn'
+        chnfile = os.path.join(self.lut_dir, fn+'.chn')
         params = self.load_chn(chnfile, coszen)
 
-        # Be careful with the two thermal values! They can only be used in 
-        # the modtran_tir functions as they require the modtran reflectivity 
-        # be set to 1 in order to use them in the RTM in radiative_transfer.py. 
+        # Be careful with the two thermal values! They can only be used in
+        # the modtran_tir functions as they require the modtran reflectivity
+        # be set to 1 in order to use them in the RTM in radiative_transfer.py.
         # Don't add these to the VSWIR functions!
         names = ['wl', 'sol', 'rhoatm', 'transm', 'sphalb', 'transup']
 
         # Don't include the thermal terms in VSWIR runs to avoid incorrect usage
-        if self.band_mode_string == 'modtran_tir':
+        if self.treat_as_emissive:
             names = names + ['thermal_upwelling', 'thermal_downwelling']
 
         results_dict = {name: param for name, param in zip(names, params)}
@@ -388,19 +440,33 @@ class ModtranRT(TabularRT):
         return results_dict
 
     def _lookup_lut(self, point):
-        ret = {}
-        for key, lut in self.luts.items():
-            ret[key] = s.array(lut(point)).ravel()
+        if np.all(np.equal(point, self.last_point_looked_up)):
+            return self.last_point_lookup_values
+        else:
+            ret = {}
+            for key, lut in self.luts.items():
+                ret[key] = np.array(lut(point)).ravel()
 
-        return ret
+            self.last_point_looked_up = point
+            self.last_point_lookup_values = ret
+            return ret
 
-    def get(self, x_RT, geom):
-        point = s.zeros((self.n_point,))
+    def get(self, x_RT: np.array, geom: Geometry) -> dict:
+        """ Get interpolated MODTRAN results at a particular location
+
+        Args:
+            x_RT: radiative-transfer portion of the statevector
+            geom: local geometry conditions for lookup
+
+        Returns:
+            interpolated modtran result
+
+        """
+        point = np.zeros((self.n_point,))
         for point_ind, name in enumerate(self.lut_grid_config):
-            if name in self.statevec:
-                ix = self.statevec.index(name)
-                x_RT_ind = self._full_to_local_statevector_position_mapping[ix]
-                point[point_ind] = x_RT[x_RT_ind]
+            if name in self.statevector_names:
+                ix = self.statevector_names.index(name)
+                point[point_ind] = x_RT[ix]
             elif name == "OBSZEN":
                 point[point_ind] = geom.OBSZEN
             elif name == "GNDALT":
@@ -426,65 +492,82 @@ class ModtranRT(TabularRT):
 
         return self._lookup_lut(point)
 
-    def get_L_atm(self, x_RT, geom):
-        if self.band_mode_string.lower() == 'modtran_vswir':
-            return self.get_L_atm_vswir(x_RT, geom)
-        elif self.band_mode_string.lower() == 'modtran_tir':
-            return self.get_L_atm_tir(x_RT, geom)
-        else:
-            raise NotImplementedError
+    def get_L_atm(self, x_RT: np.array, geom: Geometry) -> np.array:
+        """ Get the interpolated MODTRAN modeled atmospheric reflectance (aka path radiance).
 
-    def get_L_atm_vswir(self, x_RT, geom):
+        Args:
+            x_RT: radiative-transfer portion of the statevector
+            geom: local geometry conditions for lookup
+
+        Returns:
+            the interpolated MODTRAN modeled atmospheric reflectance
+
+        """
+        if self.treat_as_emissive:
+            return self._get_L_atm_tir(x_RT, geom)
+        else:
+            return self._get_L_atm_vswir(x_RT, geom)
+
+    def _get_L_atm_vswir(self, x_RT: np.array, geom: Geometry) -> np.array:
         r = self.get(x_RT, geom)
         rho = r['rhoatm']
-        rdn = rho/s.pi*(self.solar_irr*self.coszen)
+        rdn = rho/np.pi*(self.solar_irr*self.coszen)
         return rdn
 
-    def get_L_atm_tir(self, x_RT, geom):
+    def _get_L_atm_tir(self, x_RT: np.array, geom: Geometry) -> np.array:
         r = self.get(x_RT, geom)
         return r['thermal_upwelling']
 
-    def get_L_down_transmitted(self, x_RT, geom):
-        if self.band_mode_string.lower() == 'modtran_vswir':
-            return self.get_L_down_transmitted_vswir(x_RT, geom)
-        elif self.band_mode_string.lower() == 'modtran_tir':
-            return self.get_L_down_transmitted_tir(x_RT, geom)
-        else:
-            raise NotImplementedError
+    def get_L_down_transmitted(self, x_RT: np.array, geom: Geometry) -> np.array:
+        """ Get the interpolated MODTRAN downward atmospheric transmittance.
 
-    def get_L_down_transmitted_vswir(self, x_RT, geom):
+        Args:
+            x_RT: radiative-transfer portion of the statevector
+            geom: local geometry conditions for lookup
+
+        Returns:
+            The interpolated MODTRAN downward atmospheric transmittance
+        """
+
+        if self.treat_as_emissive:
+            return self._get_L_down_transmitted_tir(x_RT, geom)
+        else:
+            return self._get_L_down_transmitted_vswir(x_RT, geom)
+
+    def _get_L_down_transmitted_vswir(self, x_RT, geom):
         r = self.get(x_RT, geom)
-        rdn = (self.solar_irr*self.coszen) / s.pi * r['transm']
+        rdn = (self.solar_irr*self.coszen) / np.pi * r['transm']
         return rdn
 
-    def get_L_down_transmitted_tir(self, x_RT, geom):
+    def _get_L_down_transmitted_tir(self, x_RT, geom):
         """thermal_downwelling already includes the transmission factor. Also
         assume there is no multiple scattering for TIR.
         """
         r = self.get(x_RT, geom)
         return r['thermal_downwelling']
 
-    def get_L_up(self, x_RT, geom):
-        """Thermal emission from the ground is provided by the thermal model, 
-        so this function is a placeholder for future upgrades."""
-        return 0
+    def wl2flt(self, wavelengths: np.array, fwhms: np.array, outfile: str) -> None:
+        """Helper function to generate Gaussian distributions around the
+        center wavelengths.
 
-    def wl2flt(self, wls, fwhms, outfile):
-        """Helper function to generate Gaussian distributions around the 
-        center wavelengths."""
+        Args:
+            wavelengths: wavelength centers
+            fwhms: full width at half max
+            outfile: file to write to
+
+        """
 
-        I = None
         sigmas = fwhms/2.355
-        span = 2.0 * (wls[1]-wls[0])  # nm
+        span = 2.0 * (wavelengths[1]-wavelengths[0])  # nm
         steps = 101
 
         with open(outfile, 'w') as fout:
 
             fout.write('Nanometer data for sensor\n')
-            for wl, fwhm, sigma in zip(wls, fwhms, sigmas):
+            for wl, fwhm, sigma in zip(wavelengths, fwhms, sigmas):
 
-                ws = wl + s.linspace(-span, span, steps)
-                vs = normal.pdf(ws, wl, sigma)
+                ws = wl + np.linspace(-span, span, steps)
+                vs = scipy.stats.norm.pdf(ws, wl, sigma)
                 vs = vs/vs[int(steps/2)]
                 wns = 10000.0/(ws/1000.0)
 
